import threading
import time
import unittest
from datetime import datetime, timedelta
from unittest.mock import MagicMock

from freezegun import freeze_time

from dynatrace_extension import EndpointStatus, EndpointStatuses, Extension, Severity, StatusValue


class KillSchedulerError(Exception):
    pass


class TestSfmPerEndpont(unittest.TestCase):
    def setUp(self, extension_name=""):
        self.ext = Extension(name=extension_name)
        self.ext.logger = MagicMock()
        self.ext._running_in_sim = True
        self.ext._client = MagicMock()
        self.ext._is_fastcheck = False
        self.i = 0
        self.test_cases = None
        self.time_machine_idx = None

        self.ext.schedule(self.callback, timedelta(seconds=1))
        self.scheduler_thread = threading.Thread(target=self.scheduler_thread_fun)

    def scheduler_thread_fun(self):
        with self.assertRaises(KillSchedulerError):
            self.ext._scheduler.run()

    def tearDown(self) -> None:
        self.ext._scheduler.enter(delay=0, priority=1, action=lambda: (_ for _ in ()).throw(KillSchedulerError()))
        self.scheduler_thread.join()
        Extension._instance = None

    def run_test(self):
        self.scheduler_thread.start()
        time.sleep(0.1)

        for case in self.test_cases[: self.time_machine_idx]:
            self.single_test_iteration(case)

        if self.time_machine_idx:
            with freeze_time(datetime.now() + timedelta(hours=2), tick=True):
                for case in self.test_cases[self.time_machine_idx :]:
                    self.single_test_iteration(case)

    def single_test_iteration(self, case):
        self.ext._client.send_sfm_logs.reset_mock()
        self.ext._build_current_status()
        time.sleep(0.05)  # sleep required becuase mocked method is called in a different thread

        if case["expected"]:
            if not isinstance(case["expected"], list):
                case["expected"] = [case["expected"]]
            self.ext._client.send_sfm_logs.assert_called_once_with(case["expected"])
        else:
            self.ext._client.send_sfm_logs.assert_not_called()

        time.sleep(1)

    def callback(self):
        assert self.test_cases

        ep_status = EndpointStatuses()
        if self.i < len(self.test_cases):
            if self.test_cases[self.i]["status"]:
                ep_status.add_endpoint_status(self.test_cases[self.i]["status"])
            self.i += 1

        return ep_status

    def expected_sfm_dict(
        self,
        device_address,
        level,
        status,
        status_msg,
        status_state,
        dt_extension_config_id="",
        dt_extension_ds="python",
        dt_extension_version="",
        dt_extension_name="",
        dt_extension_config_label="",
    ):

        return {
            "device.address": device_address,
            "level": level,
            "message": f"{device_address}: [{status_state}] - {status} {status_msg}",
            "dt.extension.config.id": dt_extension_config_id,
            "dt.extension.ds": dt_extension_ds,
            "dt.extension.version": dt_extension_version,
            "dt.extension.name": dt_extension_name,
            "dt.extension.config.label": dt_extension_config_label,
        }

    def test_endpoint_sfm_ok(self):
        self.time_machine_idx = 5
        self.test_cases = [
            {"status": EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 1"), "expected": None},
            {"status": EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 1"), "expected": None},
            {"status": None, "expected": None},
            {
                "status": EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 2"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:80", level="INFO", status="OK", status_msg="All good 2", status_state="NEW"
                ),
            },
            {"status": EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 2"), "expected": None},
            # Time machine applied
            {"status": None, "expected": None},
            {
                "status": EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 3"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:80", level="INFO", status="OK", status_msg="All good 3", status_state="NEW"
                ),
            },
        ]

        self.run_test()

    def test_endpoint_sfm_nok(self):
        self.time_machine_idx = 2
        self.test_cases = [
            {
                "status": EndpointStatus("1.2.3.4:80", StatusValue.WARNING, "Warning 1"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:80",
                    level="WARN",
                    status="WARNING",
                    status_msg="Warning 1",
                    status_state="INITIAL",
                ),
            },
            {"status": EndpointStatus("1.2.3.4:80", StatusValue.WARNING, "Warning 1"), "expected": None},
            # Time machine applied
            {
                "status": None,
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:80",
                    level="WARN",
                    status="WARNING",
                    status_msg="Warning 1",
                    status_state="ONGOING",
                ),
            },
            {
                "status": EndpointStatus("1.2.3.4:80", StatusValue.GENERIC_ERROR, "Generic error 1"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:80",
                    level="ERROR",
                    status="GENERIC_ERROR",
                    status_msg="Generic error 1",
                    status_state="NEW",
                ),
            },
            {
                "status": EndpointStatus("1.2.3.4:80", StatusValue.WARNING, "Warning 1"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:80",
                    level="WARN",
                    status="WARNING",
                    status_msg="Warning 1",
                    status_state="NEW",
                ),
            },
            {
                "status": EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 1"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:80", level="INFO", status="OK", status_msg="All good 1", status_state="NEW"
                ),
            },
        ]

        self.run_test()

    def test_endpoint_sfm_severity(self):
        def expected_severtity(status: StatusValue):
            requirements = {
                StatusValue.EMPTY: Severity.INFO,
                StatusValue.OK: Severity.INFO,
                StatusValue.GENERIC_ERROR: Severity.ERROR,
                StatusValue.INVALID_ARGS_ERROR: Severity.ERROR,
                StatusValue.EEC_CONNECTION_ERROR: Severity.ERROR,
                StatusValue.INVALID_CONFIG_ERROR: Severity.ERROR,
                StatusValue.AUTHENTICATION_ERROR: Severity.ERROR,
                StatusValue.DEVICE_CONNECTION_ERROR: Severity.ERROR,
                StatusValue.WARNING: Severity.WARN,
                StatusValue.UNKNOWN_ERROR: Severity.ERROR,
            }

            for s in StatusValue:
                assert s in requirements.keys()

            return requirements[status].value

        self.test_cases = [
            {
                "status": EndpointStatus("1.2.3.4:123", status, f"Status {i}"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:123",
                    level=expected_severtity(status),
                    status=status.value,
                    status_msg=f"Status {i}",
                    status_state="INITIAL" if i == 0 else "NEW",
                ),
            }
            for i, status in enumerate(StatusValue)
        ]

        self.run_test()

    def test_endpoint_custom_blocked(self):
        self.ext = None
        Extension._instance = None
        self.setUp(extension_name="custom:custom_ext_unit_test")
        self.test_cases = [
            {"status": EndpointStatus("1.2.3.4:80", StatusValue.GENERIC_ERROR, "Generic error 1"), "expected": None}
        ]

        self.run_test()

    def test_endpoint_independent_ongoing(self):
        self.test_cases = [
            {
                "status": EndpointStatus("1.2.3.4:1", StatusValue.WARNING, "Warning 1"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:1",
                    level="WARN",
                    status="WARNING",
                    status_msg="Warning 1",
                    status_state="INITIAL",
                ),
            },
            {
                "status": EndpointStatus("1.2.3.4:2", StatusValue.WARNING, "Warning 2"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:2",
                    level="WARN",
                    status="WARNING",
                    status_msg="Warning 2",
                    status_state="INITIAL",
                ),
            },
            {
                "status": EndpointStatus("1.2.3.4:1", StatusValue.WARNING, "Warning 1"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:1",
                    level="WARN",
                    status="WARNING",
                    status_msg="Warning 1",
                    status_state="ONGOING",
                ),
            },
            {
                "status": EndpointStatus("1.2.3.4:2", StatusValue.WARNING, "Warning 2"),
                "expected": self.expected_sfm_dict(
                    device_address="1.2.3.4:2",
                    level="WARN",
                    status="WARNING",
                    status_msg="Warning 2",
                    status_state="ONGOING",
                ),
            },
            {
                "status": None,
                "expected": [
                    self.expected_sfm_dict(
                        device_address="1.2.3.4:1",
                        level="WARN",
                        status="WARNING",
                        status_msg="Warning 1",
                        status_state="ONGOING",
                    ),
                    self.expected_sfm_dict(
                        device_address="1.2.3.4:2",
                        level="WARN",
                        status="WARNING",
                        status_msg="Warning 2",
                        status_state="ONGOING",
                    ),
                ],
            },
        ]

        self.scheduler_thread.start()
        time.sleep(0.1)

        self.single_test_iteration(self.test_cases[0])

        with freeze_time(datetime.now() + timedelta(hours=1), tick=True):
            self.single_test_iteration(self.test_cases[1])

            with freeze_time(datetime.now() + timedelta(hours=1), tick=True):
                self.single_test_iteration(self.test_cases[2])

                with freeze_time(datetime.now() + timedelta(hours=1), tick=True):
                    self.single_test_iteration(self.test_cases[3])

                    with freeze_time(datetime.now() + timedelta(hours=2), tick=True):
                        self.single_test_iteration(self.test_cases[4])



"""
mureq is a replacement for python-requests, intended to be vendored
in-tree by Linux systems software and other lightweight applications.
mureq is copyright 2021 by its contributors and is released under the
0BSD ("zero-clause BSD") license.
"""

import contextlib
import io
import os.path
import socket
import ssl
import sys
import urllib.parse
from http.client import HTTPConnection, HTTPException, HTTPMessage, HTTPSConnection

__version__ = "0.2.0"

__all__ = [
    "HTTPException",
    "TooManyRedirects",
    "Response",
    "yield_response",
    "request",
    "get",
    "post",
    "head",
    "put",
    "patch",
    "delete",
]

DEFAULT_TIMEOUT = 15.0

# e.g. "Python 3.8.10"
DEFAULT_UA = "Python " + sys.version.split()[0]


def request(method, url, *, read_limit=None, **kwargs):
    """request performs an HTTP request and reads the entire response body.
    :param str method: HTTP method to request (e.g. 'GET', 'POST')
    :param str url: URL to request
    :param read_limit: maximum number of bytes to read from the body, or None for no limit
    :type read_limit: int or None
    :param kwargs: optional arguments defined by yield_response
    :return: Response object
    :rtype: Response
    :raises: HTTPException
    """
    with yield_response(method, url, **kwargs) as response:
        try:
            body = response.read(read_limit)
        except HTTPException:
            raise
        except OSError as e:
            raise HTTPException(str(e)) from e
        return Response(response.url, response.status, _prepare_incoming_headers(response.headers), body)


def get(url, **kwargs):
    """get performs an HTTP GET request."""
    return request("GET", url=url, **kwargs)


def post(url, body=None, **kwargs):
    """post performs an HTTP POST request."""
    return request("POST", url=url, body=body, **kwargs)


def head(url, **kwargs):
    """head performs an HTTP HEAD request."""
    return request("HEAD", url=url, **kwargs)


def put(url, body=None, **kwargs):
    """put performs an HTTP PUT request."""
    return request("PUT", url=url, body=body, **kwargs)


def patch(url, body=None, **kwargs):
    """patch performs an HTTP PATCH request."""
    return request("PATCH", url=url, body=body, **kwargs)


def delete(url, **kwargs):
    """delete performs an HTTP DELETE request."""
    return request("DELETE", url=url, **kwargs)


@contextlib.contextmanager
def yield_response(
    method,
    url,
    *,
    unix_socket=None,
    timeout=DEFAULT_TIMEOUT,
    headers=None,
    params=None,
    body=None,
    form=None,
    json=None,
    verify=True,
    source_address=None,
    max_redirects=None,
    ssl_context=None,
):
    """yield_response is a low-level API that exposes the actual
    http.client.HTTPResponse via a contextmanager.
    Note that unlike mureq.Response, http.client.HTTPResponse does not
    automatically canonicalize multiple appearances of the same header by
    joining them together with a comma delimiter. To retrieve canonicalized
    headers from the response, use response.getheader():
    https://docs.python.org/3/library/http.client.html#http.client.HTTPResponse.getheader
    :param str method: HTTP method to request (e.g. 'GET', 'POST')
    :param str url: URL to request
    :param unix_socket: path to Unix domain socket to query, or None for a normal TCP request
    :type unix_socket: str or None
    :param timeout: timeout in seconds, or None for no timeout (default: 15 seconds)
    :type timeout: float or None
    :param headers: HTTP headers as a mapping or list of key-value pairs
    :param params: parameters to be URL-encoded and added to the query string, as a mapping or list of key-value pairs
    :param body: payload body of the request
    :type body: bytes or None
    :param form: parameters to be form-encoded and sent as the payload body, as a mapping or list of key-value pairs
    :param json: object to be serialized as JSON and sent as the payload body
    :param bool verify: whether to verify TLS certificates (default: True)
    :param source_address: source address to bind to for TCP
    :type source_address: str or tuple(str, int) or None
    :param max_redirects: maximum number of redirects to follow, or None (the default) for no redirection
    :type max_redirects: int or None
    :param ssl_context: TLS config to control certificate validation, or None for default behavior
    :type ssl_context: ssl.SSLContext or None
    :return: http.client.HTTPResponse, yielded as context manager
    :rtype: http.client.HTTPResponse
    :raises: HTTPException
    """
    method = method.upper()
    headers = _prepare_outgoing_headers(headers)
    enc_params = _prepare_params(params)
    body = _prepare_body(body, form, json, headers)

    visited_urls = []

    while max_redirects is None or len(visited_urls) <= max_redirects:
        url, conn, path = _prepare_request(
            method,
            url,
            enc_params=enc_params,
            timeout=timeout,
            unix_socket=unix_socket,
            verify=verify,
            source_address=source_address,
            ssl_context=ssl_context,
        )
        enc_params = ""  # don't reappend enc_params if we get redirected
        visited_urls.append(url)
        try:
            try:
                conn.request(method, path, headers=headers, body=body)
                response = conn.getresponse()
            except HTTPException:
                raise
            except OSError as e:
                # wrap any IOError that is not already an HTTPException
                # in HTTPException, exposing a uniform API for remote errors
                raise HTTPException(str(e)) from e
            redirect_url = _check_redirect(url, response.status, response.headers)
            if max_redirects is None or redirect_url is None:
                response.url = url  # https://bugs.python.org/issue42062
                yield response
                return
            else:
                url = redirect_url
                if response.status == 303:
                    # 303 See Other: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/303
                    method = "GET"
        finally:
            conn.close()

    raise TooManyRedirects(visited_urls)


class Response:
    """Response contains a completely consumed HTTP response.
    :ivar str url: the retrieved URL, indicating whether a redirection occurred
    :ivar int status_code: the HTTP status code
    :ivar http.client.HTTPMessage headers: the HTTP headers
    :ivar bytes body: the payload body of the response
    """

    __slots__ = ("url", "status_code", "headers", "body")

    def __init__(self, url, status_code, headers, body):
        self.url, self.status_code, self.headers, self.body = url, status_code, headers, body

    def __repr__(self):
        return f"Response(status_code={self.status_code:d})"

    @property
    def ok(self):
        """ok returns whether the response had a successful status code
        (anything other than a 40x or 50x)."""
        return not (400 <= self.status_code < 600)

    @property
    def content(self):
        """content returns the response body (the `body` member). This is an
        alias for compatibility with requests.Response."""
        return self.body

    def raise_for_status(self):
        """raise_for_status checks the response's success code, raising an
        exception for error codes."""
        if not self.ok:
            raise HTTPErrorStatus(self.status_code)

    def json(self):
        """Attempts to deserialize the response body as UTF-8 encoded JSON."""
        import json as jsonlib

        return jsonlib.loads(self.body)

    def _debugstr(self):
        buf = io.StringIO()
        print("HTTP", self.status_code, file=buf)
        for k, v in self.headers.items():
            print(f"{k}: {v}", file=buf)
        print(file=buf)
        try:
            print(self.body.decode("utf-8"), file=buf)
        except UnicodeDecodeError:
            print(f"<{len(self.body)} bytes binary data>", file=buf)
        return buf.getvalue()


class TooManyRedirects(HTTPException):
    """TooManyRedirects is raised when automatic following of redirects was
    enabled, but the server redirected too many times without completing."""

    pass


class HTTPErrorStatus(HTTPException):
    """HTTPErrorStatus is raised by Response.raise_for_status() to indicate an
    HTTP error code (a 40x or a 50x). Note that a well-formed response with an
    error code does not result in an exception unless raise_for_status() is
    called explicitly.
    """

    def __init__(self, status_code):
        self.status_code = status_code

    def __str__(self):
        return f"HTTP response returned error code {self.status_code:d}"


# end public API, begin internal implementation details

_JSON_CONTENTTYPE = "application/json"
_FORM_CONTENTTYPE = "application/x-www-form-urlencoded"


class UnixHTTPConnection(HTTPConnection):
    """UnixHTTPConnection is a subclass of HTTPConnection that connects to a
    Unix domain stream socket instead of a TCP address.
    """

    def __init__(self, path, timeout=DEFAULT_TIMEOUT):
        super(UnixHTTPConnection, self).__init__("localhost", timeout=timeout)
        self._unix_path = path

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.settimeout(self.timeout)
            sock.connect(self._unix_path)
        except Exception:
            sock.close()
            raise
        self.sock = sock


def _check_redirect(url, status, response_headers):
    """Return the URL to redirect to, or None for no redirection."""
    if status not in (301, 302, 303, 307, 308):
        return None
    location = response_headers.get("Location")
    if not location:
        return None
    parsed_location = urllib.parse.urlparse(location)
    if parsed_location.scheme:
        # absolute URL
        return location

    old_url = urllib.parse.urlparse(url)
    if location.startswith("/"):
        # absolute path on old hostname
        return urllib.parse.urlunparse(
            (
                old_url.scheme,
                old_url.netloc,
                parsed_location.path,
                parsed_location.params,
                parsed_location.query,
                parsed_location.fragment,
            )
        )

    # relative path on old hostname
    old_dir, _old_file = os.path.split(old_url.path)
    new_path = os.path.join(old_dir, location)
    return urllib.parse.urlunparse(
        (
            old_url.scheme,
            old_url.netloc,
            new_path,
            parsed_location.params,
            parsed_location.query,
            parsed_location.fragment,
        )
    )


def _prepare_outgoing_headers(headers):
    if headers is None:
        headers = HTTPMessage()
    elif not isinstance(headers, HTTPMessage):
        new_headers = HTTPMessage()
        if hasattr(headers, "items"):
            iterator = headers.items()
        else:
            iterator = iter(headers)
        for k, v in iterator:
            new_headers[k] = v
        headers = new_headers
    _setdefault_header(headers, "User-Agent", DEFAULT_UA)
    return headers


# XXX join multi-headers together so that get(), __getitem__(),
# etc. behave intuitively, then stuff them back in an HTTPMessage.
def _prepare_incoming_headers(headers):
    headers_dict = {}
    for k, v in headers.items():
        headers_dict.setdefault(k, []).append(v)
    result = HTTPMessage()
    # note that iterating over headers_dict preserves the original
    # insertion order in all versions since Python 3.6:
    for k, vlist in headers_dict.items():
        result[k] = ",".join(vlist)
    return result


def _setdefault_header(headers, name, value):
    if name not in headers:
        headers[name] = value


def _prepare_body(body, form, json, headers):
    if body is not None:
        if not isinstance(body, bytes):
            raise TypeError("body must be bytes or None", type(body))
        return body

    if json is not None:
        _setdefault_header(headers, "Content-Type", _JSON_CONTENTTYPE)
        import json as jsonlib

        return jsonlib.dumps(json).encode("utf-8")

    if form is not None:
        _setdefault_header(headers, "Content-Type", _FORM_CONTENTTYPE)
        return urllib.parse.urlencode(form, doseq=True)

    return None


def _prepare_params(params):
    if params is None:
        return ""
    return urllib.parse.urlencode(params, doseq=True)


def _prepare_request(
    method,
    url,
    *,
    enc_params="",
    timeout=DEFAULT_TIMEOUT,
    source_address=None,
    unix_socket=None,
    verify=True,
    ssl_context=None,
):
    """Parses the URL, returns the path and the right HTTPConnection subclass."""
    parsed_url = urllib.parse.urlparse(url)

    is_unix = unix_socket is not None
    scheme = parsed_url.scheme.lower()
    if scheme.endswith("+unix"):
        scheme = scheme[:-5]
        is_unix = True
        if scheme == "https":
            raise ValueError("https+unix is not implemented")

    if scheme not in ("http", "https"):
        raise ValueError("unrecognized scheme", scheme)

    is_https = scheme == "https"
    host = parsed_url.hostname
    port = 443 if is_https else 80
    if parsed_url.port:
        port = parsed_url.port

    if is_unix and unix_socket is None:
        unix_socket = urllib.parse.unquote(parsed_url.netloc)

    path = parsed_url.path
    if parsed_url.query:
        if enc_params:
            path = f"{path}?{parsed_url.query}&{enc_params}"
        else:
            path = f"{path}?{parsed_url.query}"
    else:
        if enc_params:
            path = f"{path}?{enc_params}"
        else:
            pass  # just parsed_url.path in this case

    if isinstance(source_address, str):
        source_address = (source_address, 0)

    if is_unix:
        conn = UnixHTTPConnection(unix_socket, timeout=timeout)
    elif is_https:
        if ssl_context is None:
            ssl_context = ssl.create_default_context()
            if not verify:
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE
        conn = HTTPSConnection(host, port, source_address=source_address, timeout=timeout, context=ssl_context)
    else:
        conn = HTTPConnection(host, port, source_address=source_address, timeout=timeout)

    munged_url = urllib.parse.urlunparse(
        (parsed_url.scheme, parsed_url.netloc, path, parsed_url.params, "", parsed_url.fragment)
    )
    return munged_url, conn, path



import threading
import time
import unittest
from datetime import timedelta
from unittest.mock import MagicMock

from dynatrace_extension import EndpointStatus, EndpointStatuses, Extension, MultiStatus, Status, StatusValue
from dynatrace_extension.sdk.communication import DebugClient


class TestStatus(unittest.TestCase):
    def tearDown(self) -> None:
        Extension._instance = None

    def test_status(self):
        status = Status(StatusValue.OK, "status message")
        self.assertEqual(status.status, StatusValue.OK)
        self.assertEqual(status.message, "status message")
        self.assertEqual(status.status.value, "OK")

    def test_overall_status(self):
        def callback():
            return 1

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False
        ext.schedule(callback, timedelta(seconds=1))
        status = ext._build_current_status()

        self.assertEqual(status.status, StatusValue.OK)
        self.assertEqual(status.message, "")

    def test_bad_status(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def bad_method():
            msg = "something went wrong"
            raise Exception(msg)

        ext.schedule(bad_method, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
        self.assertIn("something went wrong", status.message)

    def test_multiple_bad_status(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def bad_method_1():
            msg = "something went wrong"
            raise Exception(msg)

        def bad_method_2():
            msg = "something broke"
            raise Exception(msg)

        ext.schedule(bad_method_1, timedelta(seconds=1))
        ext.schedule(bad_method_2, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
        self.assertIn("something went wrong", status.message)
        self.assertIn("something broke", status.message)

    def test_callback_taking_too_long_sets_status(self):
        extension = Extension()
        extension.logger = MagicMock()
        extension._running_in_sim = True
        extension._is_fastcheck = False
        extension._client = MagicMock()

        def callback():
            time.sleep(1)

        extension.schedule(callback, timedelta(seconds=1))
        extension._scheduler.run(blocking=False)
        time.sleep(2)

        self.assertTrue(extension._scheduled_callbacks[1].status.is_error())
        self.assertIn("longer than the interval", extension._scheduled_callbacks[1].status.message)

    def test_direct_status_return(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            return Status(StatusValue.OK, "foo1")

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)
        self.assertIn("foo1", status.message)

    def test_direct_statuses_return(self):
        def callback():
            return Status(StatusValue.OK, "foo1")

        def custom_query():
            return Status(StatusValue.EMPTY, "foo2")

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback, timedelta(seconds=1))
        ext.schedule(custom_query, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)
        self.assertIn("foo1", status.message)
        self.assertIn("foo2", status.message)

    def test_multistatus_empty(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            ret = MultiStatus()
            return ret

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)

    def test_multistatus_ok(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            ret = MultiStatus()
            ret.add_status(StatusValue.OK, "foo1")
            ret.add_status(StatusValue.OK, "foo2")
            return ret

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)
        self.assertIn("foo1", status.message)
        self.assertIn("foo2", status.message)

    def test_multistatus_warning_1(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            ret = MultiStatus()
            ret.add_status(StatusValue.OK, "foo1")
            ret.add_status(StatusValue.GENERIC_ERROR, "foo2")
            return ret

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn("foo1", status.message)
        self.assertIn("foo2", status.message)

    def test_multistatus_warning_2(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            ret = MultiStatus()
            ret.add_status(StatusValue.OK, "foo1")
            ret.add_status(StatusValue.WARNING, "foo2")
            return ret

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn("foo1", status.message)
        self.assertIn("foo2", status.message)

    def test_multistatus_warning_3(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            ret = MultiStatus()
            ret.add_status(StatusValue.GENERIC_ERROR, "foo1")
            ret.add_status(StatusValue.WARNING, "foo2")
            return ret

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn("foo1", status.message)
        self.assertIn("foo2", status.message)

    def test_multistatus_single_warning(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            ret = MultiStatus()
            ret.add_status(StatusValue.WARNING, "foo1")
            return ret

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn("foo1", status.message)

    def test_multistatus_error(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            ret = MultiStatus()
            ret.add_status(StatusValue.AUTHENTICATION_ERROR, "foo1")
            ret.add_status(StatusValue.DEVICE_CONNECTION_ERROR, "foo2")
            return ret

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(1)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
        self.assertIn("foo1", status.message)
        self.assertIn("foo2", status.message)

    def test_endpoint_status_all_ok(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            statuses = EndpointStatuses()
            statuses.add_endpoint_status(EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 1"))
            statuses.add_endpoint_status(EndpointStatus("4.5.6.7:80", StatusValue.OK, "All good 2"))

            statuses.add_endpoint_status(EndpointStatus("6.7.8.9:80", StatusValue.OK, "All good 3"))

            return statuses

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)
        self.assertIn("OK: 3 NOK: 0", status.message)

    def test_endpoint_status_some_faulty_endpoints(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            statuses = EndpointStatuses()
            statuses.add_endpoint_status(EndpointStatus("1.2.3.4:80", StatusValue.OK, "Invalid authorization scheme 1"))
            statuses.add_endpoint_status(
                EndpointStatus("4.5.6.7:80", StatusValue.DEVICE_CONNECTION_ERROR, "Invalid authorization scheme 2")
            )

            statuses.add_endpoint_status(
                EndpointStatus("6.7.8.9:80", StatusValue.DEVICE_CONNECTION_ERROR, "Invalid authorization scheme 3")
            )

            return statuses

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn(
            "OK: 1 NOK: 2 NOK_reported_errors: 4.5.6.7:80 - DEVICE_CONNECTION_ERROR "
            "Invalid authorization scheme 2, 6.7.8.9:80 - DEVICE_CONNECTION_ERROR Invalid authorization scheme 3",
            status.message,
        )

    def test_endpoint_status_all_faulty_endpoints(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            statuses = EndpointStatuses()
            statuses.add_endpoint_status(
                EndpointStatus("1.2.3.4:80", StatusValue.AUTHENTICATION_ERROR, "Invalid authorization scheme 4")
            )
            statuses.add_endpoint_status(
                EndpointStatus("4.5.6.7:80", StatusValue.DEVICE_CONNECTION_ERROR, "Invalid authorization scheme 5")
            )

            statuses.add_endpoint_status(
                EndpointStatus("6.7.8.9:80", StatusValue.DEVICE_CONNECTION_ERROR, "Invalid authorization scheme 6")
            )

            return statuses

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
        self.assertIn(
            "OK: 0 NOK: 3 NOK_reported_errors: 1.2.3.4:80 - AUTHENTICATION_ERROR "
            "Invalid authorization scheme 4, 4.5.6.7:80 - DEVICE_CONNECTION_ERROR "
            "Invalid authorization scheme 5, 6.7.8.9:80 - DEVICE_CONNECTION_ERROR Invalid authorization scheme 6",
            status.message,
        )

    def test_endpoint_empty(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            statuses = EndpointStatuses()
            return statuses

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)
        self.assertNotIn("OK: 0 NOK: 0", status.message)

    def test_endpoint_status_single_warning(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            statuses = EndpointStatuses()
            statuses.add_endpoint_status(
                EndpointStatus("1.2.3.4:80", StatusValue.WARNING, "Invalid authorization scheme A")
            )
            return statuses

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn(
            "OK: 0 NOK: 1 NOK_reported_errors: 1.2.3.4:80 - WARNING Invalid authorization scheme A", status.message
        )

    def test_endpoint_status_single_error(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        def callback():
            statuses = EndpointStatuses()
            statuses.add_endpoint_status(
                EndpointStatus("1.2.3.4:80", StatusValue.AUTHENTICATION_ERROR, "Invalid authorization scheme")
            )
            return statuses

        ext.schedule(callback, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
        self.assertIn(
            "OK: 0 NOK: 1 NOK_reported_errors: 1.2.3.4:80 - AUTHENTICATION_ERROR Invalid authorization scheme",
            status.message,
        )

    def test_endpoint_merge_ok(self):
        def callback_ep_status_1():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("1.2.3.4:80", StatusValue.OK, "All good 1"))
            return status

        def callback_ep_status_2():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("5.6.7.8:90", StatusValue.OK, "All good 2"))
            status.add_endpoint_status(EndpointStatus("10.11.12.13:100", StatusValue.OK, "All good 3"))
            return status

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback_ep_status_1, timedelta(seconds=1))
        ext.schedule(callback_ep_status_2, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)
        self.assertIn("OK: 3 NOK: 0", status.message)

    def test_endpoint_merge_error(self):
        def callback_ep_status_1():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT_1", StatusValue.AUTHENTICATION_ERROR, "EP1 MSG"))
            return status

        def callback_ep_status_2():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT_2", StatusValue.INVALID_CONFIG_ERROR, "EP2 MSG"))
            return status

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback_ep_status_1, timedelta(seconds=1))
        ext.schedule(callback_ep_status_2, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
        self.assertIn(
            "OK: 0 NOK: 2 NOK_reported_errors: EP_HINT_1 - AUTHENTICATION_ERROR EP1 MSG, EP_HINT_2 - INVALID_CONFIG_ERROR EP2 MSG",
            status.message,
        )

    def test_endpoint_merge_warning(self):
        def callback_ep_status_1():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT_1", StatusValue.OK, "EP1 MSG"))
            return status

        def callback_ep_status_2():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT_2", StatusValue.WARNING, "EP2 MSG"))
            return status

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback_ep_status_1, timedelta(seconds=1))
        ext.schedule(callback_ep_status_2, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn("OK: 1 NOK: 1 NOK_reported_errors: EP_HINT_2 - WARNING EP2 MSG", status.message)

    def test_overall_status_error(self):
        def callback_ep_status():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT", StatusValue.UNKNOWN_ERROR, "EP MSG"))
            return status

        def callback_multistatus():
            status = MultiStatus()
            status.add_status(StatusValue.INVALID_ARGS_ERROR, "MULTI MSG")
            return status

        def callback_status():
            status = Status(StatusValue.EEC_CONNECTION_ERROR, "STATUS MSG")
            return status

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback_ep_status, timedelta(seconds=1))
        ext.schedule(callback_multistatus, timedelta(seconds=1))
        ext.schedule(callback_status, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
        self.assertIn(
            "Endpoints OK: 0 NOK: 1 NOK_reported_errors: EP_HINT - UNKNOWN_ERROR EP MSG"
            "\ncallback_multistatus: GENERIC_ERROR - MULTI MSG\ncallback_status: EEC_CONNECTION_ERROR - STATUS MSG",
            status.message,
        )

    def test_overall_status_ok(self):
        def callback_ep_status():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT", StatusValue.OK, "EP MSG"))
            return status

        def callback_multistatus():
            status = MultiStatus()
            status.add_status(StatusValue.OK, "")
            return status

        def callback_status():
            status = Status(StatusValue.OK, "STATUS MSG")
            return status

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback_ep_status, timedelta(seconds=1))
        ext.schedule(callback_multistatus, timedelta(seconds=1))
        ext.schedule(callback_status, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.OK)
        self.assertIn("Endpoints OK: 1 NOK: 0\ncallback_status: OK - STATUS MSG", status.message)

    def test_overall_status_warning_1(self):
        def callback_ep_status():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT", StatusValue.WARNING, "EP MSG"))
            return status

        def callback_multistatus():
            status = MultiStatus()
            status.add_status(StatusValue.OK, "MULTI MSG")
            return status

        def callback_status():
            status = Status(StatusValue.OK, "")
            return status

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback_ep_status, timedelta(seconds=1))
        ext.schedule(callback_multistatus, timedelta(seconds=1))
        ext.schedule(callback_status, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn(
            "Endpoints OK: 0 NOK: 1 NOK_reported_errors: EP_HINT - WARNING EP MSG\ncallback_multistatus: OK - MULTI MSG",
            status.message,
        )

    def test_overall_status_warning_2(self):
        def callback_ep_status():
            status = EndpointStatuses()
            status.add_endpoint_status(EndpointStatus("EP_HINT", StatusValue.INVALID_CONFIG_ERROR, ""))
            return status

        def callback_multistatus():
            status = MultiStatus()
            status.add_status(StatusValue.INVALID_ARGS_ERROR, "")
            return status

        def callback_status():
            status = Status(StatusValue.WARNING, "")
            return status

        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        ext.schedule(callback_ep_status, timedelta(seconds=1))
        ext.schedule(callback_multistatus, timedelta(seconds=1))
        ext.schedule(callback_status, timedelta(seconds=1))
        ext._scheduler.run(blocking=False)
        time.sleep(0.01)

        status = ext._build_current_status()
        self.assertEqual(status.status, StatusValue.WARNING)
        self.assertIn(
            "Endpoints OK: 0 NOK: 1 NOK_reported_errors: EP_HINT - INVALID_CONFIG_ERROR "
            "\ncallback_multistatus: GENERIC_ERROR - \ncallback_status: WARNING - ",
            status.message,
        )

    def test_endpoint_status_skipped_interval(self):
        ext = Extension()
        ext.logger = MagicMock()
        ext._running_in_sim = True
        ext._client = DebugClient("", "", MagicMock())
        ext._is_fastcheck = False

        skipped_callback_call_counter = 0
        regular_callback_call_counter = 0

        def skipped_callback():
            nonlocal skipped_callback_call_counter
            skipped_callback_call_counter += 1

            statuses = EndpointStatuses()
            statuses.add_endpoint_status(
                EndpointStatus("skipped_callback", StatusValue.GENERIC_ERROR, "skipped_callback_msg")
            )
            return statuses

        def regular_callback():
            nonlocal regular_callback_call_counter
            regular_callback_call_counter += 1
            statuses = EndpointStatuses()
            statuses.add_endpoint_status(
                EndpointStatus("regular_callback", StatusValue.UNKNOWN_ERROR, "regular_callback_msg")
            )
            return statuses

        ext.schedule(skipped_callback, timedelta(seconds=10))  # called only once during test
        ext.schedule(regular_callback, timedelta(seconds=1))

        # Runngin scheduler in another thread as we need it to run in parallel in this test
        class KillSchedulerError(Exception):
            pass

        def scheduler_thread_impl(ext: Extension):
            try:
                ext._scheduler.run(blocking=True)
            except KillSchedulerError:
                pass

        scheduler_thread = threading.Thread(target=scheduler_thread_impl, args=(ext,))
        scheduler_thread.start()
        time.sleep(0.01)

        # 5 second of test
        for _ in range(5):
            status = ext._build_current_status()
            self.assertEqual(status.status, StatusValue.GENERIC_ERROR)
            self.assertIn(
                (
                    "Endpoints OK: 0 NOK: 2 NOK_reported_errors: "
                    "skipped_callback - GENERIC_ERROR skipped_callback_msg, regular_callback - UNKNOWN_ERROR regular_callback_msg"
                ),
                status.message,
            )
            time.sleep(1)

        ext._scheduler.enter(delay=0, priority=1, action=lambda: (_ for _ in ()).throw(KillSchedulerError()))
        scheduler_thread.join()

        # Confirm schedulered called callbacks as requested
        self.assertEqual(skipped_callback_call_counter, 1)
        self.assertEqual(regular_callback_call_counter, 6)



# SPDX-FileCopyrightText: 2023-present Dynatrace LLC
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import json
import logging
import sys
from abc import ABC, abstractmethod
from collections.abc import Generator, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TypeVar

from .status import Status
from .vendor.mureq.mureq import HTTPException, Response, request

CONTENT_TYPE_JSON = "application/json;charset=utf-8"
CONTENT_TYPE_PLAIN = "text/plain;charset=utf-8"
COUNT_METRIC_ITEMS_DICT = TypeVar("COUNT_METRIC_ITEMS_DICT", str, list[str])

# TODO - I believe these can be adjusted via RuntimeConfig, they can't be constants
MAX_MINT_LINES_PER_REQUEST = 1000
MAX_LOG_EVENTS_PER_REQUEST = 50_000
MAX_LOG_REQUEST_SIZE = 5_000_000  # actually 5_242_880
MAX_METRIC_REQUEST_SIZE = 1_000_000  # actually 1_048_576

HTTP_BAD_REQUEST = 400


class CommunicationClient(ABC):
    """
    Abstract class for extension communication
    """

    @abstractmethod
    def get_activation_config(self) -> dict:
        pass

    @abstractmethod
    def get_extension_config(self) -> str:
        pass

    @abstractmethod
    def get_feature_sets(self) -> dict[str, list[str]]:
        pass

    @abstractmethod
    def register_count_metrics(self, pattern: dict[str, dict[str, COUNT_METRIC_ITEMS_DICT]]) -> None:
        pass

    @abstractmethod
    def send_count_delta_signal(self, metric_keys: set[str]) -> None:
        pass

    @abstractmethod
    def send_status(self, status: Status) -> dict:
        pass

    @abstractmethod
    def send_keep_alive(self) -> str:
        pass

    @abstractmethod
    def send_metrics(self, mint_lines: list[str]) -> list[MintResponse]:
        pass

    @abstractmethod
    def send_events(self, event: dict | list[dict], eec_enrichment: bool) -> list[dict | None]:
        pass

    @abstractmethod
    def send_sfm_metrics(self, metrics: list[str]) -> MintResponse:
        pass

    @abstractmethod
    def get_cluster_time_diff(self) -> int:
        pass

    @abstractmethod
    def send_dt_event(self, event: dict) -> None:
        pass

    @abstractmethod
    def send_sfm_logs(self, sfm_logs: dict | list[dict]) -> list[dict | None]:
        pass


class HttpClient(CommunicationClient):
    """
    Concrete implementation of the client, this one handles the communication with the EEC
    """

    def __init__(self, base_url: str, datasource_id: str, id_token_file_path: str, logger: logging.Logger):
        self._activation_config_url = f"{base_url}/userconfig/{datasource_id}"
        self._extension_config_url = f"{base_url}/extconfig/{datasource_id}"
        self._metric_url = f"{base_url}/mint/{datasource_id}"
        self._sfm_url = f"{base_url}/sfm/{datasource_id}"
        self._sfm_logs_url = f"{base_url}/sfmlogs/{datasource_id}"
        self._keep_alive_url = f"{base_url}/alive/{datasource_id}"
        self._timediff_url = f"{base_url}/timediffms"
        self._events_url = f"{base_url}/logs/{datasource_id}"
        self._count_metric_register_url = f"{base_url}/countmetricregister/{datasource_id}"
        self._count_delta_signal_url = f"{base_url}/countmetricdeltasignal/{datasource_id}"
        self._feature_sets_query = "?feature_sets_json"
        self._event_ingest_url = f"{base_url}/events/{datasource_id}"

        with open(id_token_file_path) as f:
            id_token = f.read()
            self._headers = {"Authorization": f"Api-Token {id_token}"}

        self.logger = logger

    def _make_request(
        self,
        url: str,
        method: str = "GET",
        body: Any = None,
        extra_headers: dict | None = None,
        is_delta_signal: bool = False,
    ) -> Response:
        if extra_headers is None:
            extra_headers = {}
        headers = {**self._headers, **extra_headers}

        response = request(method, url, body=body, headers=headers)
        self.logger.debug(f"Response from {url}: {response}")
        if response.status_code >= HTTP_BAD_REQUEST:
            if not is_delta_signal:
                self.logger.warning(f"Error HTTP {response.status_code} from {url}: {response.content}")
        return response

    def get_activation_config(self) -> dict:
        try:
            response = self._make_request(self._activation_config_url, "GET")
        except HTTPException as err:
            self.logger.error(f"HTTP exception: {err}")
            return {}

        if response.status_code < HTTP_BAD_REQUEST:
            try:
                return response.json()
            except Exception as err:
                self.logger.error(f"JSON parse failure: {err}")
            return {}
        else:
            self.logger.error(f"Can't get activation configuration ({response.content}). Extension is stopped.")
            sys.exit(1)

    def get_extension_config(self) -> str:
        try:
            response = self._make_request(self._extension_config_url, "GET")
            return response.content.decode("utf-8")
        except HTTPException as err:
            self.logger.error(f"HTTP exception: {err}")
            return ""

    def get_feature_sets(self) -> dict[str, list[str]]:
        try:
            response = self._make_request(self._extension_config_url + self._feature_sets_query, "GET")
        except HTTPException as err:
            self.logger.error(f"HTTP exception: {err}")
            return {}

        if response.status_code < HTTP_BAD_REQUEST:
            try:
                return response.json()
            except Exception as err:
                self.logger.error(f"JSON parse failure: {err}")
                return {}

        return {}

    def register_count_metrics(self, json_pattern: dict[str, dict[str, COUNT_METRIC_ITEMS_DICT]]) -> None:
        register_data = json.dumps(json_pattern).encode("utf-8")
        try:
            response = self._make_request(
                self._count_metric_register_url,
                "POST",
                register_data,
                extra_headers={"Content-Type": CONTENT_TYPE_JSON},
            )
            if response.ok:
                self.logger.debug(
                    f"Monotonic cache converter successful registration for metric {list(json_pattern.keys())}."
                )
        except HTTPException:
            self.logger.error(
                f"Monotonic cache converter registration request error for metric {list(json_pattern.keys())}."
            )

    def send_count_delta_signal(self, metric_keys: set[str]) -> None:
        json_data = {"metric_keys": list(metric_keys), "filter_dimensions": {}}
        delta_signal_data = json.dumps(json_data).encode("utf-8")
        try:
            response = self._make_request(
                self._count_delta_signal_url,
                "POST",
                delta_signal_data,
                extra_headers={"Content-Type": CONTENT_TYPE_JSON},
                is_delta_signal=True,
            )
            if response.ok:
                self.logger.debug(
                    f"Monotonic converter cache delta calculation signal success for metric {metric_keys}."
                )
            else:
                self.logger.debug(
                    f"Not enough metrics of type {metric_keys} cached in monotonic cache converter to calculate delta."
                )
        except HTTPException:
            self.logger.error(
                f"Monotonic cache converter delta calculation signal request error for metric {metric_keys}."
            )

    def send_dt_event(self, event: dict[str, str | int | dict[str, str]]):
        json_data = json.dumps(event).encode("utf-8")
        try:
            response = self._make_request(
                self._event_ingest_url, "POST", json_data, extra_headers={"Content-Type": CONTENT_TYPE_JSON}
            )
            if response.ok:
                self.logger.debug(f"DT Event sent to EEC, content: {json_data.decode('utf-8')}")
            else:
                self.logger.debug(f"DT Event request failed: {response.content}")
        except HTTPException:
            self.logger.error(f"DT Event request HTTP exception, request body: {json_data.decode('utf-8')}")

    def send_status(self, status: Status) -> dict:
        encoded_data = json.dumps(status.to_json()).encode("utf-8")
        self.logger.debug(f"Sending status to EEC: {status}")
        response = self._make_request(
            self._keep_alive_url, "POST", encoded_data, extra_headers={"Content-Type": CONTENT_TYPE_JSON}
        ).content
        return json.loads(response.decode("utf-8"))

    def send_keep_alive(self):
        return self.send_status(Status())

    def send_metrics(self, mint_lines: list[str]) -> list[MintResponse]:
        responses = []

        # We divide into batches of MAX_METRIC_REQUEST_SIZE bytes to avoid hitting the body size limit
        batches = divide_into_batches(mint_lines, MAX_METRIC_REQUEST_SIZE, "\n")
        for batch in batches:
            response = self._make_request(
                self._metric_url, "POST", batch, extra_headers={"Content-Type": CONTENT_TYPE_PLAIN}
            ).json()
            self.logger.debug(f"{self._metric_url}: {response}")
            mint_response = MintResponse.from_json(response)
            responses.append(mint_response)
        return responses

    def send_events(self, events: dict | list[dict], eec_enrichment: bool = True) -> list[dict | None]:
        self.logger.debug(f"Sending log events: {events}")
        return self._send_events(self._events_url, events, eec_enrichment)

    def send_sfm_logs(self, sfm_logs: dict | list[dict]):
        self.logger.debug(f"Sending SFM logs: {sfm_logs}")
        return self._send_events(self._sfm_logs_url, sfm_logs)

    def _send_events(self, url, events: dict | list[dict], eec_enrichment: bool = True) -> list[dict | None]:
        responses = []
        if isinstance(events, dict):
            events = [events]
        batches = divide_into_batches(events, MAX_LOG_REQUEST_SIZE)

        for batch in batches:
            try:
                eec_response = self._make_request(
                    url,
                    "POST",
                    batch,
                    extra_headers={"Content-Type": CONTENT_TYPE_JSON, "eec-enrichment": str(eec_enrichment).lower()},
                ).json()
                responses.append(eec_response)
            except json.JSONDecodeError:
                responses.append(None)

        return responses

    def send_sfm_metrics(self, mint_lines: list[str]) -> MintResponse:
        mint_data = "\n".join(mint_lines).encode("utf-8")
        return MintResponse.from_json(
            self._make_request(
                self._sfm_url, "POST", mint_data, extra_headers={"Content-Type": CONTENT_TYPE_PLAIN}
            ).json()
        )

    def get_cluster_time_diff(self) -> int:
        response = self._make_request(self._timediff_url, "GET")
        time_diff = response.json()["clusterDiffMs"]
        return time_diff


class DebugClient(CommunicationClient):
    """
    This client is used for debugging purposes
    It does not send metrics to Dynatrace, but prints them to the console
    """

    def __init__(
        self,
        activation_config_path: str,
        extension_config_path: str,
        logger: logging.Logger,
        secrets_path: str = "secrets.json",
        local_ingest: bool = False,
        local_ingest_port: int = 14499,
        print_metrics: bool = True,
    ):
        self.secrets = {}
        if secrets_path and Path(secrets_path).exists():
            with open(secrets_path) as f:
                self.secrets = json.load(f)

        self.activation_config = {}
        if activation_config_path and Path(activation_config_path).exists():
            with open(activation_config_path) as f:
                raw_activation_config = f.read()
                self.activation_config = json.loads(
                    self.replace_secrets_in_activation_config(self.secrets, raw_activation_config)
                )

        self.extension_config = ""
        if not extension_config_path:
            extension_config_path = "extension/extension.yaml"
        if Path(extension_config_path).exists():
            with open(extension_config_path) as f:
                self.extension_config = f.read()
        self.logger = logger
        self.local_ingest = local_ingest
        self.local_ingest_port = local_ingest_port
        self.print_metrics = print_metrics

    def get_activation_config(self) -> dict:
        return self.activation_config

    def get_extension_config(self) -> str:
        return self.extension_config

    def get_feature_sets(self) -> dict[str, list[str]]:
        # This is only called from dt-sdk run, where PyYaml is installed because of dt-cli
        # Do NOT move this to the top of the file
        import yaml  # type: ignore

        # Grab the feature sets from the extension.yaml file
        extension_yaml = yaml.safe_load(self.extension_config)
        if not extension_yaml:
            return {}

        yaml_feature_sets = extension_yaml.get("python", {}).get("featureSets", [])
        if not yaml_feature_sets:
            return {}

        # Construct the object that the SDK expects
        feature_sets = {}
        for feature_set in yaml_feature_sets:
            feature_set_name = feature_set["featureSet"]
            if feature_set_name in self.activation_config.get("featureSets", []):
                feature_sets[feature_set_name] = [metric["key"] for metric in feature_set["metrics"]]

        return feature_sets

    def register_count_metrics(self, pattern: dict[str, dict[str, COUNT_METRIC_ITEMS_DICT]]) -> None:
        self.logger.info(f"Registering metrics in converter: {pattern}")

    def send_count_delta_signal(self, metric_keys: set[str]) -> None:
        self.logger.info(f"Sending delta signal for: {metric_keys}")

    def send_dt_event(self, event: dict) -> None:
        self.logger.info(f"Sending DT Event: {event}")

    def send_status(self, status: Status) -> dict:
        self.logger.info(f"send_status: '{status}'")
        return {}

    def send_keep_alive(self):
        return self.send_status(Status())

    def send_metrics(self, mint_lines: list[str]) -> list[MintResponse]:
        total_lines = len(mint_lines)
        self.logger.info(f"Start sending {total_lines} metrics to the EEC")

        responses = []

        batches = divide_into_batches(mint_lines, MAX_METRIC_REQUEST_SIZE)
        for batch in batches:
            lines = json.loads(batch)
            if self.local_ingest:
                response = request(
                    "POST",
                    f"http://localhost:{self.local_ingest_port}/metrics/ingest",
                    body="\n".join(lines).encode(),
                    headers={"Content-Type": CONTENT_TYPE_PLAIN},
                ).json()
                mint_response = MintResponse.from_json(response)
                responses.append(mint_response)
            elif self.print_metrics:
                for line in lines:
                    self.logger.info(f"send_metric: {line}")

        return responses

    def send_events(self, events: dict | list[dict], eec_enrichment: bool = True) -> list[dict | None]:
        self.logger.info(f"send_events (enrichment = {eec_enrichment}): {len(events)} events")
        if isinstance(events, dict):
            events = [events]
        if self.print_metrics:
            for event in events:
                self.logger.info(f"send_event: {event}")
        return []

    def send_sfm_metrics(self, mint_lines: list[str]) -> MintResponse:
        for line in mint_lines:
            self.logger.info(f"send_sfm_metric: {line}")
        return MintResponse(lines_invalid=0, lines_ok=len(mint_lines), error=None, warnings=None)

    def get_cluster_time_diff(self) -> int:
        return 0

    def replace_secrets_in_activation_config(self, secrets: dict, activation_config_string: str) -> str:
        for secret_name, secret_value in secrets.items():
            activation_config_string = activation_config_string.replace(f"{{{{{secret_name}}}}}", str(secret_value))

        return activation_config_string

    def send_sfm_logs(self, sfm_logs: dict | list[dict]) -> list[dict | None]:
        if isinstance(sfm_logs, dict):
            sfm_logs = [sfm_logs]

        self.logger.info(f"send_sfm_logs: {len(sfm_logs)} logs")

        if self.print_metrics:
            for log in sfm_logs:
                self.logger.info(f"send_sfm_log: {log}")
        return []


def divide_into_batches(
    items: Sequence[dict | str], max_size_bytes: int, join_with: str | None = None
) -> Generator[bytes, None, None]:
    """
    Yield successive batches from a list, according to sizing limitations

    :param items: The list items to divide, they myst be encodable to bytes
    :param max_size_bytes: The maximum size of the payload in bytes
    :param join_with: A string to join the items with before encoding
    :return: A generator of batches of log events already encoded
    """

    if not items:
        return

    if join_with is not None:
        joined = join_with.join(items)  # type: ignore
        encoded = f"{joined}".encode(errors="replace")
    else:
        encoded = json.dumps(items).encode(errors="replace")
    size = len(encoded)
    if size <= max_size_bytes:
        yield encoded
        return

    # if we get here, the payload is too large, split it in half until we have chunks that are small enough
    half = len(items) // 2
    first_half = items[:half]
    second_half = items[half:]
    yield from divide_into_batches(first_half, max_size_bytes, join_with)
    yield from divide_into_batches(second_half, max_size_bytes, join_with)


@dataclass
class MintResponse:
    lines_ok: int
    lines_invalid: int
    error: dict | None
    warnings: dict | None

    @staticmethod
    def from_json(json_data: dict) -> MintResponse:
        return MintResponse(
            lines_ok=json_data.get("linesOk", 0),
            lines_invalid=json_data.get("linesInvalid", 0),
            error=json_data.get("error"),
            warnings=json_data.get("warnings"),
        )

    def __str__(self) -> str:
        return f"MintResponse(lines_ok={self.lines_ok}, lines_invalid={self.lines_invalid}, error={self.error}, warnings={self.warnings})"



# SPDX-FileCopyrightText: 2023-present Dynatrace LLC
#
# SPDX-License-Identifier: MIT

# Ignore F401 because these are not used here, but are used by extension developers
# ruff: noqa: F401

from .sdk.activation import ActivationConfig, ActivationType
from .sdk.event import Severity
from .sdk.extension import DtEventType, Extension
from .sdk.helper import (
    get_activation_config,
    get_helper_extension,
    report_dt_event,
    report_dt_event_dict,
    report_event,
    report_log_event,
    report_log_events,
    report_log_lines,
    report_metric,
    report_mint_lines,
    run_extension,
    schedule,
    schedule_function,
)
from .sdk.metric import Metric, MetricType, SummaryStat
from .sdk.status import EndpointStatus, EndpointStatuses, IgnoreStatus, MultiStatus, Status, StatusValue



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path

PREFIX_HOST = "HOST"
PREFIX_PG = "PROCESS_GROUP"
PREFIX_PGI = "PROCESS_GROUP_INSTANCE"


@dataclass
class EntryProperties:
    technologies: list[str]
    pg_technologies: list[str]
    extra_properties: dict[str, str]

    @staticmethod
    def from_json(json_data: dict[str, str]) -> EntryProperties:
        technologies = json_data.get("Technologies", "").split(",")
        pg_technologies = json_data.get("pgTechnologies", "").split(",")
        extra_properties = {k: v for k, v in json_data.items() if k not in ["Technologies", "pgTechnologies"]}
        return EntryProperties(technologies, pg_technologies, extra_properties)


@dataclass
class PortBinding:
    ip: str
    port: int

    @staticmethod
    def from_string(data: str) -> PortBinding:
        ip, port = data.split("_")
        return PortBinding(ip, int(port))


@dataclass
class ProcessProperties:
    cmd_line: str | None
    exe_path: str | None
    parent_pid: int | None
    work_dir: str | None
    listening_ports: list[int]
    port_bindings: list[PortBinding]
    docker_mount: str | None
    docker_container_id: str | None
    listening_internal_ports: str | None

    @staticmethod
    def from_json(json_data: dict) -> ProcessProperties:
        cmd_line = json_data.get("CmdLine")
        exe_path = json_data.get("ExePath")
        parent_pid = int(json_data.get("ParentPid", -1))
        work_dir = json_data.get("WorkDir")
        listening_ports = [int(p) for p in json_data.get("ListeningPorts", "").split(" ") if p != ""]
        port_bindings = [PortBinding.from_string(p) for p in json_data.get("PortBindings", "").split(";") if p != ""]
        docker_mount = json_data.get("DockerMount")
        docker_container_id = json_data.get("DockerContainerId")
        listening_internal_ports = json_data.get("ListeningInternalPorts")
        return ProcessProperties(
            cmd_line,
            exe_path,
            parent_pid,
            work_dir,
            listening_ports,
            port_bindings,
            docker_mount,
            docker_container_id,
            listening_internal_ports,
        )


@dataclass
class Process:
    pid: int
    process_name: str
    properties: ProcessProperties

    @staticmethod
    def from_json(json_data: dict) -> Process:
        pid = int(json_data.get("pid", "-1"))
        process_name = json_data.get("process_name", "unknown-process-name")
        all_properties = {}
        for p in json_data.get("properties", []):
            all_properties.update(p)
        properties = ProcessProperties.from_json(all_properties)
        return Process(pid, process_name, properties)


@dataclass
class Entry:
    group_id: str
    node_id: str
    group_instance_id: str
    process_type: int
    group_name: str
    processes: list[Process]
    properties: EntryProperties

    @staticmethod
    def from_json(json_data: dict) -> Entry:
        group_id = json_data.get("group_id", "0X0000000000000000")
        group_id = f"{PREFIX_PG}-{group_id[-16:]}"

        node_id = json_data.get("node_id", "0X0000000000000000")

        group_instance_id = json_data.get("group_instance_id", "0X0000000000000000")
        group_instance_id = f"{PREFIX_PGI}-{group_instance_id[-16:]}"

        process_type = int(json_data.get("process_type", "0"))
        group_name = json_data.get("group_name", "unknown-group-name")
        processes = [Process.from_json(p) for p in json_data.get("processes", [])]

        # The structure here was never thought out, so we have to check for both keys and merge them into one object
        properties_list: list[dict[str, str]] = json_data.get("properties", [])
        technologies = []
        pg_technologies = []
        # There may be other useful properties included such as: mssql_instance_name.
        extra_properties = {}

        for prop in properties_list:
            for key, value in prop.items():
                match key:
                    case "Technologies":
                        technologies.extend(value.split(","))
                    case "pgTechnologies":
                        pg_technologies.extend(value.split(","))
                    case _:
                        extra_properties[key] = value

        properties = EntryProperties(technologies, pg_technologies, extra_properties)

        return Entry(group_id, node_id, group_instance_id, process_type, group_name, processes, properties)


@dataclass
class Snapshot:
    host_id: str
    entries: list[Entry]

    @staticmethod
    def parse_from_file(snapshot_file: Path | str | None = None) -> Snapshot:
        """Returns a process snapshot object like EF1.0 used to do"""

        if snapshot_file is None:
            snapshot_file = find_log_dir() / "plugin" / "oneagent_latest_snapshot.log"

        with open(snapshot_file) as f:
            snapshot_json = json.load(f)

        host_id = snapshot_json.get("host_id", "0X0000000000000000")
        host_id = f"{PREFIX_HOST}-{host_id[-16:]}"
        entries = [Entry.from_json(e) for e in snapshot_json.get("entries", [])]
        return Snapshot(host_id, entries)

    # Returns list of Process groups matching a technology. Use to simulate activation
    def get_process_groups_by_technology(self, technology: str) -> list[Entry]:
        pgs = []
        for entry in self.entries:
            if technology in entry.properties.technologies:
                pgs.append(entry)

        return pgs


def find_config_directory() -> Path:
    """
    Attempt to find the OneAgent config directory.
    Note, the user can never modify these directories
    Windows -> https://docs.dynatrace.com/docs/shortlink/oneagent-disk-requirements-windows#oneagent-files-aging-mechanism
    Linux -> https://docs.dynatrace.com/docs/shortlink/oneagent-disk-requirements-linux#sizes
    """
    config_dir_base = os.path.expandvars("%PROGRAMDATA%") if os.name == "nt" else "/var/lib"
    config_dir = Path(config_dir_base) / "dynatrace" / "oneagent" / "agent" / "config"
    if config_dir.exists():
        return config_dir
    file_path = Path(__file__).resolve()

    while file_path.parent != file_path:
        file_path = file_path.parent
        if file_path.name == "agent":
            return file_path / "config"

    msg = "Could not find the OneAgent config directory"
    raise Exception(msg)


def find_log_dir() -> Path:
    """
    Attempt to find the OneAgent log directory.
    This is always stored in the installation.conf file.
    So we attempt to find the installation.conf file and read the LogDir property
    Returns: the Path to the log directory
    """
    config_dir = find_config_directory()
    installation_conf = config_dir / "installation.conf"
    if not installation_conf.exists():
        msg = f"Could not find installation.conf at {installation_conf}"
        raise Exception(msg)

    with open(installation_conf) as f:
        for line in f:
            if line.startswith("LogDir"):
                log_dir = line.split("=")[1].strip()
                if not log_dir:
                    return (
                        Path(os.path.expandvars("%PROGRAMDATA%") + "/dynatrace/oneagent/log")
                        if os.name == "nt"
                        else Path("/var/log/dynatrace/oneagent")
                    )
                return Path(log_dir)
    msg = f"Could not find LogDir in {installation_conf}"
    raise Exception(msg)



import os
import re
import shutil
from pathlib import Path
from typing import Final, NamedTuple

output_mode_folder: Final = 0o755
output_mode_file: Final = 0o644
template_base_folder: Final = Path(__file__).parent


class ReplaceString(NamedTuple):
    what: str
    replace: str


def replace_placeholders(file: Path, replaces: list[tuple[str, str]]):
    with open(file) as f:
        contents = f.read()
    for replace in replaces:
        old_text, new_text = replace
        contents = contents.replace(old_text, new_text)
    with open(file, "w") as f:
        f.write(contents)


def copy_templates(source: Path, destination: Path, extension_name: str):
    extension_name_lower = extension_name.lower()
    extension_name_capitalize = extension_name.capitalize()
    extension_name_dash = extension_name_lower.replace("_", "-")
    extension_prefix = "custom:"
    replaces = [
        ("%Extension_Name%", extension_name_capitalize),
        ("%extension_name%", extension_name_lower),
        ("%extension-name%", extension_name_dash),
        ("%extension-prefix%", extension_prefix),
    ]
    for file in source.iterdir():
        if file.is_dir() and file.name != "__pycache__":
            # If it is a folder, recursively copy and process the template files
            dir_name = file.name
            if dir_name == "extension_name":
                dir_name = extension_name
            dest_dir = Path(destination, dir_name)
            dest_dir.mkdir(mode=output_mode_folder, exist_ok=True)
            copy_templates(file, dest_dir, extension_name)

        elif file.is_file() and file.name.endswith(".template"):
            # If it is a file, copy and process it
            file_name = file.name.removesuffix(".template")
            dest_file = Path(destination, file_name)
            shutil.copy(file, dest_file)
            os.chmod(dest_file, output_mode_file)
            replace_placeholders(dest_file, replaces)


def is_pep8_compliant(extension_name: str) -> bool:
    """
    Check if extension name is valid according to PEP-8
    https://peps.python.org/pep-0008/#package-and-module-names
    :param extension_name: extension name to check
    :return: True if extension name is valid, False otherwise
    """
    if not re.match("[a-z][a-z0-9]*(_[a-z0-9]+)*$", extension_name):
        return False
    return True


def generate_extension(extension_name: str, output: Path) -> Path:
    output.mkdir(mode=output_mode_folder, exist_ok=True, parents=True)
    try:
        copy_templates(template_base_folder / "extension_template", output, extension_name)
        return output.resolve()
    except Exception:
        shutil.rmtree(output)
        raise



import json
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from threading import Lock

from .event import Severity


class DynatraceDeprecatedError(Exception):
    pass


class StatusValue(Enum):
    EMPTY = ""
    OK = "OK"
    GENERIC_ERROR = "GENERIC_ERROR"
    INVALID_ARGS_ERROR = "INVALID_ARGS_ERROR"
    EEC_CONNECTION_ERROR = "EEC_CONNECTION_ERROR"
    INVALID_CONFIG_ERROR = "INVALID_CONFIG_ERROR"
    AUTHENTICATION_ERROR = "AUTHENTICATION_ERROR"
    DEVICE_CONNECTION_ERROR = "DEVICE_CONNECTION_ERROR"
    WARNING = "WARNING"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"

    def is_error(self) -> bool:
        # WARNING is treated as an error
        return self not in (StatusValue.OK, StatusValue.EMPTY)

    def is_warning(self) -> bool:
        return self == StatusValue.WARNING


class IgnoreStatus:
    pass


class Status:
    def __init__(self, status: StatusValue = StatusValue.EMPTY, message: str = "", timestamp: int | None = None):
        self.status = status
        self.message = message
        self.timestamp = timestamp

    def to_json(self) -> dict:
        status = {"status": self.status.value, "message": self.message}
        if self.timestamp:
            status["timestamp"] = self.timestamp  # type: ignore
        return status

    def __repr__(self):
        return json.dumps(self.to_json())

    def is_error(self) -> bool:
        # WARNING is treated as an error
        return self.status.is_error()

    def is_warning(self) -> bool:
        return self.status.is_warning()


class MultiStatus:
    def __init__(self) -> None:
        self.statuses: list[Status] = []

    def add_status(self, status: StatusValue, message):
        self.statuses.append(Status(status, message))

    def build(self) -> Status:
        ret = Status(StatusValue.OK)
        if len(self.statuses) == 0:
            return ret

        messages = []
        all_ok = True
        all_err = True
        any_warning = False

        for stored_status in self.statuses:
            if stored_status.message != "":
                messages.append(stored_status.message)

            if stored_status.is_warning():
                any_warning = True

            if stored_status.is_error():
                all_ok = False
            else:
                all_err = False

        ret.message = ", ".join(messages)

        if any_warning:
            ret.status = StatusValue.WARNING
        elif all_ok:
            ret.status = StatusValue.OK
        elif all_err:
            ret.status = StatusValue.GENERIC_ERROR
        else:
            ret.status = StatusValue.WARNING

        return ret


class EndpointStatus:
    def __init__(self, endpoint_hint: str, short_status: StatusValue, message: str | None = None):
        self.endpoint = endpoint_hint
        self.status: StatusValue = short_status
        self.message = message

    def __repr__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        return isinstance(other, EndpointStatus) and self.__dict__ == other.__dict__


class EndpointStatuses:
    def __init__(self, total_endpoints_number=None) -> None:
        if total_endpoints_number is not None:
            msg = (
                "EndpointStatuses.__init__: usage of `total_endpoints_number` parameter is abandoned. "
                "Use other class methods to explicitly report all status changes for any endpoint."
            )
            raise DynatraceDeprecatedError(msg)

        self._lock = Lock()
        self._endpoints_statuses: dict[str, EndpointStatus] = {}

    def add_endpoint_status(self, status: EndpointStatus):
        with self._lock:
            self._endpoints_statuses[status.endpoint] = status


class StatusState(Enum):
    INITIAL = "INITIAL"
    NEW = "NEW"
    ONGOING = "ONGOING"


@dataclass
class EndpointStatusRecord:
    ep_status: EndpointStatus
    last_sent: datetime | None
    state: StatusState

    def __repr__(self):
        return str(self.__dict__)


class EndpointStatusesMap:
    RESENDING_INTERVAL = timedelta(hours=2)

    def __init__(self, send_sfm_logs_function: Callable) -> None:
        self._lock = Lock()
        self._ep_records: dict[str, EndpointStatusRecord] = {}
        self._send_sfm_logs_function = send_sfm_logs_function
        self._logs_to_send: list[str] = []

    def contains_any_status(self) -> bool:
        return len(self._ep_records) > 0

    def update_ep_statuses(self, new_ep_statuses: EndpointStatuses):
        with self._lock:
            with new_ep_statuses._lock:
                for endpoint, ep_status in new_ep_statuses._endpoints_statuses.items():
                    if endpoint not in self._ep_records.keys():
                        self._ep_records[endpoint] = EndpointStatusRecord(
                            ep_status=ep_status, last_sent=None, state=StatusState.INITIAL
                        )
                    elif ep_status != self._ep_records[endpoint].ep_status:
                        self._ep_records[endpoint] = EndpointStatusRecord(
                            ep_status=ep_status, last_sent=None, state=StatusState.NEW
                        )

    def send_ep_logs(self):
        logs_to_send = []

        with self._lock:
            for ep_record in self._ep_records.values():
                if self._should_be_reported(ep_record):
                    logs_to_send.append(
                        self._prepare_ep_status_log(
                            ep_record.ep_status.endpoint,
                            ep_record.state,
                            ep_record.ep_status.status,
                            ep_record.ep_status.message,
                        )
                    )
                    ep_record.last_sent = datetime.now()
                    ep_record.state = StatusState.ONGOING

        if logs_to_send:
            self._send_sfm_logs_function(logs_to_send)

    def _should_be_reported(self, ep_record: EndpointStatusRecord):
        if ep_record.ep_status.status == StatusValue.OK:
            return ep_record.state == StatusState.NEW
        elif ep_record.state in (StatusState.INITIAL, StatusState.NEW):
            return True
        elif ep_record.state == StatusState.ONGOING and (
            ep_record.last_sent is None or datetime.now() - ep_record.last_sent >= self.RESENDING_INTERVAL
        ):
            return True
        else:
            return False

    def _prepare_ep_status_log(
        self, endpoint_name: str, prefix: StatusState, status_value: StatusValue, status_message: str
    ) -> dict:
        level = Severity.ERROR.value

        if status_value.is_error() is False:
            level = Severity.INFO.value
        elif status_value.is_warning():
            level = Severity.WARN.value

        ep_status_log = {
            "device.address": endpoint_name,
            "level": level,
            "message": f"{endpoint_name}: [{prefix.value}] - {status_value.value} {status_message}",
        }

        return ep_status_log

    def build_common_status(self) -> Status:
        with self._lock:
            # Summarize all statuses
            ok_count = 0
            nok_count = 0
            error_messages = []
            has_warning_status = False

            for ep_record in self._ep_records.values():
                ep_status = ep_record.ep_status
                if ep_status.status.is_warning():
                    has_warning_status = True

                if ep_status.status.is_error():
                    nok_count += 1
                    error_messages.append(f"{ep_status.endpoint} - {ep_status.status.value} {ep_status.message}")
                else:
                    ok_count += 1

            # Early return if all OK
            if nok_count == 0:
                return Status(StatusValue.OK, f"Endpoints OK: {ok_count} NOK: 0")

            # Build final status if some errors present
            common_msg = ", ".join(error_messages)
            all_endpoints_faulty = ok_count == 0

            if all_endpoints_faulty and not has_warning_status:
                status_value = StatusValue.GENERIC_ERROR
            else:
                status_value = StatusValue.WARNING

            message = f"Endpoints OK: {ok_count} NOK: {nok_count} NOK_reported_errors: {common_msg}"
            return Status(status=status_value, message=message)



import os
import shutil
import stat
import subprocess
import sys
from pathlib import Path

import typer
from dtcli.server_api import upload as dt_cli_upload  # type: ignore
from dtcli.server_api import validate as dt_cli_validate
from rich.console import Console

from ..__about__ import __version__
from .create import generate_extension, is_pep8_compliant
from .schema import ExtensionYaml

app = typer.Typer(pretty_exceptions_show_locals=False, pretty_exceptions_enable=False)
console = Console()

# if we are not python 3.10.X, exit with an error
if sys.version_info < (3, 10) or sys.version_info >= (3, 11):
    console.print(f"Python 3.10.X is required to build extensions, you are using {sys.version_info}", style="bold red")
    sys.exit(1)

CERT_DIR_ENVIRONMENT_VAR = "DT_CERTIFICATES_FOLDER"
CERTIFICATE_DEFAULT_PATH = Path.home() / ".dynatrace" / "certificates"


# show version
@app.command()
def version():
    """
    Show the version of the CLI
    """
    console.print(f"dt-extensions-sdk version {__version__}", style="bold green")


@app.command()
def run(
    extension_dir: Path = typer.Argument("."),
    activation_config: str = "activation.json",
    secrets: str = "secrets.json",
    fast_check: bool = typer.Option(False, "--fastcheck"),
    local_ingest: bool = typer.Option(False, "--local-ingest"),
    local_ingest_port: int = typer.Option(14499, "--local-ingest-port"),
    print_metrics: bool = typer.Option(True),
):
    """
    Runs an extension, this is used during development to locally run and test an extension

    :param extension_dir: The directory of the extension, by default this is the current directory
    :param activation_config: The activation config file, by default this is activation.json
    :param secrets: The secrets file to be used to enrich the activation config, by default this is secrets.json
    :param fast_check: If true, run a fastcheck and exits
    :param local_ingest: If true, send metrics to localhost:14499 on top of printing them
    :param local_ingest_port: The port to send metrics to, by default this is 14499
    :param print_metrics: If true, print metrics to the console
    """

    # This parses the yaml, which validates it before running
    extension_yaml = ExtensionYaml(extension_dir / "extension/extension.yaml")
    try:
        command = [
            sys.executable,
            "-m",
            extension_yaml.python.runtime.module,
            "--activationconfig",
            activation_config,
            "--secrets",
            secrets,
        ]
        if fast_check:
            command.append("--fastcheck")
        if local_ingest:
            command.append("--local-ingest")
            command.append(f"--local-ingest-port={local_ingest_port}")
        if not print_metrics:
            command.append("--no-print-metrics")
        run_process(command, cwd=extension_dir)
    except KeyboardInterrupt:
        console.print("\nRun interrupted with a KeyboardInterrupt, stopping", style="bold yellow")


@app.command(help="Runs wheel, assemble and sign. Downloads dependencies, creates and signs the extension zip file")
def build(
    extension_dir: Path = typer.Argument(Path("."), help="Path to the python extension"),
    private_key: Path = typer.Option(
        Path(CERTIFICATE_DEFAULT_PATH) / "developer.pem",
        "--private-key",
        "-k",
        help="Path to the dev fused key-certificate",
    ),
    target_directory: Path | None = typer.Option(None, "--target-directory", "-t"),
    extra_platforms: list[str] | None = typer.Option(
        None, "--extra-platform", "-e", help="Download wheels for an extra platform"
    ),
    extra_index_url: str | None = typer.Option(
        None, "--extra-index-url", "-i", help="Extra index url to use when downloading dependencies"
    ),
    find_links: str | None = typer.Option(
        None, "--find-links", "-f", help="Extra index url to use when downloading dependencies"
    ),
    only_extra_platforms: bool | None = typer.Option(
        False,
        "--only-extra-platforms",
        "-o",
        help="Only build for the extra platforms, useful when building from arm64 (mac)",
    ),
):
    """
    Builds and signs an extension using the developer fused key-certificate

    :param extension_dir: The directory of the extension, by default this is the current directory
    :param private_key: The path to the developer fused key-certificate, if not specified, we try to locate the file developer.pem under the environment
    variable DT_CERTIFICATES_FOLDER
    :param target_directory: The directory to put the extension zip file in, if not specified, we put it in the dist
    folder
    :param extra_platforms: Attempt to also download wheels for an extra platform (e.g. manylinux1_x86_64 or win_amd64)
    :param extra_index_url: Extra index url to use when downloading dependencies
    :param find_links: Extra index url to use when downloading dependencies
    :param only_extra_platforms: If true, only build for the extra platforms, useful when building from arm64
    """
    console.print(f"Building and signing extension from {extension_dir} to {target_directory}", style="cyan")
    if target_directory is None:
        target_directory = extension_dir / "dist"
    if not target_directory.exists():
        target_directory.mkdir()

    console.print("Stage 1 - Download and build dependencies", style="bold blue")
    wheel(extension_dir, extra_platforms, extra_index_url, find_links, only_extra_platforms)

    console.print("Stage 2 - Create the extension zip file", style="bold blue")
    built_zip = assemble(extension_dir, target_directory)

    console.print("Stage 3 - Sign the extension", style="bold blue")
    extension_yaml = ExtensionYaml(Path(extension_dir) / "extension" / "extension.yaml")
    output = target_directory / extension_yaml.zip_file_name()
    sign(built_zip, private_key, output)

    console.print(f"Stage 4 - Delete {built_zip}", style="bold blue")
    built_zip.unlink()


@app.command(help="Creates the extension zip file, without signing it yet")
def assemble(
    extension_dir: Path = typer.Argument(".", help="Path to the python extension"),
    output: Path = typer.Option(None, "--output", "-o"),
    force: bool = typer.Option(True, "--force", "-f", help="Force overwriting the output zip file"),
) -> Path:
    """
    Creates the extension zip file (not yet signed)

    :param extension_dir: The directory of the extension, by default this is the current directory
    :param output: The path to the output zip file, if not specified, we put it in the dist folder
    :param force: If true, overwrite the output zip file if it exists
    """

    # This checks if the yaml is valid, because it parses it
    # Also validates that the schema files are valid and exist
    extension_yaml = ExtensionYaml(Path(extension_dir) / "extension" / "extension.yaml")
    extension_yaml.validate()

    # Checks that the module name is valid and exists in the filesystem
    module_folder = Path(extension_dir) / extension_yaml.python.runtime.module
    src_module_folder = Path("src") / module_folder
    if not module_folder.exists() and not src_module_folder.exists():
        msg = f"Extension module folder {module_folder} not found"
        raise FileNotFoundError(msg)

    # This is the zip file that will contain the extension
    if output is None:
        dist_dir = Path(extension_dir) / "dist"
        if not dist_dir.exists():
            dist_dir.mkdir()
        output = dist_dir / "extension.zip"
    elif output.exists() and output.is_dir():
        output = output / "extension.zip"

    command = ["dt", "ext", "assemble", "--source", f"{Path(extension_dir) / 'extension'}", "--output", f"{output}"]
    if force:
        command.append("--force")
    run_process(command)
    console.print(f"Built the extension zip file to {output}", style="bold green")
    return output


@app.command(help="Downloads the dependencies of the extension to the lib folder")
def wheel(
    extension_dir: Path = typer.Argument(".", help="Path to the python extension"),
    extra_platforms: list[str] | None = typer.Option(
        None, "--extra-platform", "-e", help="Download wheels for an extra platform"
    ),
    extra_index_url: str | None = typer.Option(
        None, "--extra-index-url", "-i", help="Extra index url to use when downloading dependencies"
    ),
    find_links: str | None = typer.Option(
        None, "--find-links", "-f", help="Extra index url to use when downloading dependencies"
    ),
    only_extra_platforms: bool | None = typer.Option(
        False,
        "--only-extra-platforms",
        "-o",
        help="Only build for the extra platforms, useful when building from arm64 (mac)",
    ),
):
    """
    Builds the extension and it's dependencies into wheel files
    Places these files in the lib folder

    :param extension_dir: The directory of the extension, by default this is the current directory
    :param extra_platforms: Attempt to also download wheels for an extra platform (e.g. manylinux1_x86_64 or win_amd64)
    :param extra_index_url: Extra index url to use when downloading dependencies
    :param find_links: Extra index url to use when downloading dependencies
    :param only_extra_platforms: If true, only build for the extra platforms, useful when building from arm64
    """
    relative_lib_folder_dir = "extension/lib"
    lib_folder: Path = extension_dir / relative_lib_folder_dir
    _clean_directory(lib_folder)

    console.print(f"Downloading dependencies to {lib_folder}", style="cyan")

    # Downloads the dependencies and places them in the lib folder
    command = [sys.executable, "-m", "pip", "wheel", "-w", relative_lib_folder_dir]
    if extra_index_url is not None:
        command.extend(["--extra-index-url", extra_index_url])
    if find_links is not None:
        command.extend(["--find-links", find_links])
    if only_extra_platforms:
        command.append("--no-deps")
    command.append(".")
    run_process(command, cwd=extension_dir)

    if extra_platforms:
        for extra_platform in extra_platforms:
            console.print(f"Downloading wheels for platform {extra_platform}", style="cyan")
            command = [
                sys.executable,
                "-m",
                "pip",
                "download",
                "-d",
                relative_lib_folder_dir,
                "--only-binary=:all:",
                "--platform",
                extra_platform,
            ]
            if extra_index_url:
                command.extend(["--extra-index-url", extra_index_url])
            if find_links:
                command.extend(["--find-links", find_links])
            command.append(".")

            run_process(command, cwd=extension_dir)

    console.print(f"Installed dependencies to {lib_folder}", style="bold green")


@app.command()
def sign(
    zip_file: Path = typer.Argument(Path("dist/extension.zip"), help="Path to the extension zip file"),
    certificate: Path = typer.Option(
        Path(CERTIFICATE_DEFAULT_PATH) / "developer.pem",
        "--certificate",
        "-c",
        help="Path to the dev fused key-certificate",
    ),
    output: Path = typer.Option(None, "--output", "-o"),
    force: bool = typer.Option(True, "--force", "-f", help="Force overwriting the output zip file"),
):
    """
    Signs the extension zip file using the provided fused key-certificate

    :param zip_file: The path to the extension zip file to sign
    :param certificate: The developer fused key-certificate to use for signing
    :param output: The path to the output zip file, if not specified, we put it in the dist folder
    :param force: If true, overwrite the output zip file if it exists
    """
    if not certificate:
        # If the user doesn't specify a certificate, we try to find it in the default directory
        # This directory can be set by the environment variable DT_CERTIFICATES_FOLDER
        if CERT_DIR_ENVIRONMENT_VAR not in os.environ:
            console.print(
                f"{CERT_DIR_ENVIRONMENT_VAR} not found in environment variables. Using {CERTIFICATE_DEFAULT_PATH} instead.",
                style="yellow",
            )
        certificate = _find_certificate(CERTIFICATE_DEFAULT_PATH)
    else:
        certificate = _find_certificate(Path(certificate))

    combined_cert_and_key = certificate

    if output is None:
        output = zip_file.parent / f"signed_{zip_file.name}"

    console.print(f"Signing file {zip_file} to {output} with certificate {certificate}", style="cyan")
    command = [
        "dt",
        "ext",
        "sign",
        "--src",
        f"{zip_file}",
        "--output",
        f"{output}",
        "--key",
        f"{combined_cert_and_key}",
    ]

    if force:
        command.append("--force")
    run_process(command)
    console.print(f"Created signed extension file {output}", style="bold green")


@app.command(help="Upload the extension to a Dynatrace environment")
def upload(
    extension_path: Path = typer.Argument(None, help="Path to the extension folder or built zip file"),
    tenant_url: str = typer.Option(None, "--tenant-url", "-u", help="Dynatrace tenant URL"),
    api_token: str = typer.Option(None, "--api-token", "-t", help="Dynatrace API token"),
    validate: bool = typer.Option(None, "--validate", "-v", help="Validate only"),
):
    """
    Uploads the extension to a Dynatrace environment

    :param extension_path: The path to the extension folder or built zip file
    :param tenant_url: The Dynatrace tenant URL
    :param api_token: The Dynatrace API token
    :param validate: If true, only validate the extension and exit
    """

    zip_file_path = extension_path
    if extension_path is None:
        extension_path = Path(".")
    else:
        extension_path = Path(extension_path)
    if extension_path.is_dir():
        yaml_path = Path(extension_path, "extension", "extension.yaml")
        extension_yaml = ExtensionYaml(yaml_path)
        zip_file_name = extension_yaml.zip_file_name()
        zip_file_path = Path(extension_path, "dist", zip_file_name)

    api_url = tenant_url or os.environ.get("DT_API_URL", "")
    api_url = api_url.rstrip("/")

    if not api_url:
        console.print("Set the --tenant-url parameter or the DT_API_URL environment variable", style="bold red")
        sys.exit(1)

    api_token = api_token or os.environ.get("DT_API_TOKEN", "")
    if not api_token:
        console.print("Set the --api-token parameter or the DT_API_TOKEN environment variable", style="bold red")
        sys.exit(1)

    if validate:
        dt_cli_validate(f"{zip_file_path}", api_url, api_token)
    else:
        dt_cli_upload(f"{zip_file_path}", api_url, api_token)
        console.print(f"Extension {zip_file_path} uploaded to {api_url}", style="bold green")


@app.command(help="Generate root and developer certificates and key")
def gencerts(
    output: Path = typer.Option(CERTIFICATE_DEFAULT_PATH, "--output", "-o", help="Path to the output directory"),
    force: bool = typer.Option(False, "--force", "-f", help="Force overwriting the certificates"),
):
    developer_pem = output / "developer.pem"
    command_gen_ca = [
        "dt",
        "ext",
        "genca",
        "--ca-cert",
        f"{output / 'ca.pem'}",
        "--ca-key",
        f"{output / 'ca.key'}",
        "--no-ca-passphrase",
    ]

    command_gen_dev_pem = [
        "dt",
        "ext",
        "generate-developer-pem",
        "--output",
        f"{developer_pem}",
        "--name",
        "Acme",
        "--ca-crt",
        f"{output / 'ca.pem'}",
        "--ca-key",
        f"{output / 'ca.key'}",
    ]

    if output.exists():
        if developer_pem.exists() and force:
            command_gen_ca.append("--force")
            developer_pem.chmod(stat.S_IREAD | stat.S_IWRITE)
            developer_pem.unlink(missing_ok=True)
        elif developer_pem.exists() and not force:
            msg = f"Certificates were NOT generated! {developer_pem} already exists. Use --force option to overwrite the certificates"
            console.print(msg, style="bold red")
            sys.exit(1)
    else:
        output.mkdir(parents=True)

    run_process(command_gen_ca)
    run_process(command_gen_dev_pem)


@app.command(help="Creates a new python extension")
def create(extension_name: str, output: Path = typer.Option(None, "--output", "-o")):
    """
    Creates a new python extension

    :param extension_name: The name of the extension
    :param output: The path to the output directory, if not specified, we will use the extension name
    """

    if not is_pep8_compliant(extension_name):
        msg = f"Extension name {extension_name} is not valid, should be short, all-lowercase and underscores can be used if it improves readability"
        raise Exception(msg)

    if output is None:
        output = Path.cwd() / extension_name
    else:
        output = output / extension_name
    extension_path = generate_extension(extension_name, output)
    console.print(f"Extension created at {extension_path}", style="bold green")


@app.command(
    "format",
    help="Runs ruff format on the extension code",
)
def fmt(extension_dir: Path = typer.Argument(".", help="Path to the python extension")):
    """
    Runs ruff format on the extension code

    :param extension_dir: The directory of the extension, by default this is the current directory
    """
    run_process(["ruff", "format", str(extension_dir.resolve())])


@app.command(help="Runs ruff check on the extension code")
def lint(
    extension_dir: Path = typer.Argument(".", help="Path to the python extension"),
    fix: bool = typer.Option(False, "--fix", "-f", help="Fix linting issues"),
):
    """
    Runs ruff lint on the extension code

    :param extension_dir: The directory of the extension, by default this is the current directory
    :param fix: If true, ask ruff to also fix the linting issues
    """
    command = ["ruff", "check", "--exit-zero"]
    if fix:
        command.append("--fix")
    command.append(str(extension_dir.resolve()))
    run_process(command)


@app.command(help="Adds ruff rules if they don't exist already")
def ruff_init(extension_dir: Path = typer.Argument(".", help="Path to the python extension")):
    """
    Adds ruff rules if they don't exist already

    :param extension_dir: The directory of the extension, by default this is the current directory
    """

    if (extension_dir / "ruff.toml").exists():
        console.print(f"ruff.toml already exists in {extension_dir.resolve()}, skipping", style="bold yellow")
        return
    else:
        # create\extension_template\ruff.toml.template -> extension\ruff.toml
        ruff_template = Path(__file__).parent / "create" / "extension_template" / "ruff.toml.template"
        ruff_file = extension_dir / "ruff.toml"
        shutil.copy(ruff_template, ruff_file)
        console.print(f"Added ruff.toml to {extension_dir.resolve()}", style="bold green")


def run_process(command: list[str], cwd: Path | None = None, env: dict | None = None, print_message: str | None = None):
    friendly_command = " ".join(command)
    if print_message is not None:
        console.print(print_message, style="cyan")
    else:
        console.print(f"Running: {friendly_command}", style="cyan")
    return subprocess.run(command, cwd=cwd, env=env, check=True)  # noqa: S603


def _clean_directory(directory: Path):
    if directory.exists():
        console.print(f"Cleaning {directory}", style="cyan")
        shutil.rmtree(directory)


def _find_certificate(path: Path) -> Path:
    """
    Verifies the existence of the file in given path or returns the default file
    """

    # If the user specified the path as a directory, we try to find developer.pem in this directory
    if path.is_dir():
        certificate = path / "developer.pem"
    else:
        certificate = path

    if not certificate.exists():
        msg = f"Certificate {certificate} not found"
        raise FileNotFoundError(msg)
    return certificate


if __name__ == "__main__":
    app()



import unittest
from pathlib import Path
from unittest.mock import MagicMock

from dynatrace_extension import Extension
from dynatrace_extension.sdk.snapshot import Snapshot

test_data_dir = Path(__file__).parent.parent / "data"


class TestSnapshot(unittest.TestCase):
    def test_extension_get_snapshot(self):
        extension = Extension()
        extension.logger = MagicMock()
        extension._running_in_sim = True
        extension.initialize()

        snapshot = extension.get_snapshot(test_data_dir / "snapshot.json")
        self.assertIsNotNone(snapshot)
        assert snapshot.host_id == "HOST-524E3E2974F9AC2A"
        self.assertEqual(len(snapshot.entries), 24)

        processes = snapshot.get_process_groups_by_technology("DOCKER")
        self.assertEqual(len(processes), 4)

    def test_snapshot_parsing(self):
        snapshot = Snapshot.parse_from_file(test_data_dir / "snapshot.json")
        self.assertIsNotNone(snapshot)
        assert snapshot.host_id == "HOST-524E3E2974F9AC2A"
        self.assertEqual(len(snapshot.entries), 24)

        for entry in snapshot.entries:
            for process in entry.processes:
                if process.process_name in ("squid", "squid.exe"):
                    self.assertEqual(entry.process_type, 0)
                    self.assertEqual(entry.group_name, "squid")
                    self.assertEqual(len(entry.properties.technologies), 1)
                    self.assertEqual(len(entry.properties.pg_technologies), 1)
                    self.assertEqual(len(entry.processes), 1)
                    self.assertEqual(process.process_name, "squid")
                    self.assertEqual(process.pid, 2245656)
                    self.assertEqual(len(process.properties.listening_ports), 1)
                    self.assertEqual(process.properties.listening_ports[0], 3128)
                    self.assertEqual(process.properties.cmd_line, "-f /etc/squid/squid.conf -NYC")
                    self.assertEqual(len(process.properties.port_bindings), 1)

                    self.assertEqual(process.properties.port_bindings[0].ip, "127.0.0.1")
                    self.assertEqual(process.properties.port_bindings[0].port, 3128)



from unittest import TestCase

from dynatrace_extension.cli.create import is_pep8_compliant


class TestTemplates(TestCase):
    def test_extension_name(self):
        self.assertFalse(is_pep8_compliant("test-ext"))
        self.assertFalse(is_pep8_compliant("_testext"))
        self.assertFalse(is_pep8_compliant("testext_"))
        self.assertFalse(is_pep8_compliant("test-ext"))
        self.assertFalse(is_pep8_compliant("TestExt"))
        self.assertFalse(is_pep8_compliant("0TestExt"))
        self.assertFalse(is_pep8_compliant("0_test_ext"))

        self.assertTrue(is_pep8_compliant("testext"))
        self.assertTrue(is_pep8_compliant("test_ext"))
        self.assertTrue(is_pep8_compliant("test_ext_name"))
        self.assertTrue(is_pep8_compliant("test1_e2xt_name4"))



import shutil
from pathlib import Path
from unittest import TestCase
from unittest.mock import NonCallableMock, call, mock_open, patch

import dynatrace_extension.cli.main as dt_sdk

SAMPLE_EXTENSION_DATA = """
name: test_extension
version: 1.0.0
"""


class TestDtSdk(TestCase):
    def setUp(self):
        self.temp_dir = Path("test_temp")

    def tearDown(self):
        shutil.rmtree(self.temp_dir, ignore_errors=True)
        pass

    @patch("dynatrace_extension.cli.main.dt_cli_upload")
    @patch("dynatrace_extension.cli.main.dt_cli_validate")
    @patch("builtins.open", mock_open(read_data=SAMPLE_EXTENSION_DATA))
    def test_dt_sdk_upload(self, mock_upload, mock_validate):
        extension_path = Path("test_extension_dir")
        tenant_url = "test_tenant_url"
        api_token = "test_api_token"

        dt_sdk.upload(extension_path, tenant_url, api_token, validate=False)

    @patch("dynatrace_extension.cli.main.subprocess.run")
    def test_dt_sdk_gen_certs(self, mock_subprocess_run: NonCallableMock):
        output = Path("certificates")

        dt_sdk.gencerts(output, force=True)
        mock_subprocess_run.assert_has_calls(
            [
                call(
                    [
                        "dt",
                        "ext",
                        "genca",
                        "--ca-cert",
                        f"{output / 'ca.pem'}",
                        "--ca-key",
                        f"{output / 'ca.key'}",
                        "--no-ca-passphrase",
                    ],
                    cwd=None,
                    env=None,
                    check=True,
                ),
                call(
                    [
                        "dt",
                        "ext",
                        "generate-developer-pem",
                        "--output",
                        f"{output / 'developer.pem'}",
                        "--name",
                        "Acme",
                        "--ca-crt",
                        f"{output / 'ca.pem'}",
                        "--ca-key",
                        f"{output / 'ca.key'}",
                    ],
                    cwd=None,
                    env=None,
                    check=True,
                ),
            ]
        )

    def test_dt_sdk_workflow(self):
        # Generate real certificates
        dt_sdk.gencerts(self.temp_dir / "certificates", force=True)

        # Check that the certificate files were created
        self.assertTrue((self.temp_dir / "certificates" / "ca.key").exists())
        self.assertTrue((self.temp_dir / "certificates" / "ca.pem").exists())
        self.assertTrue((self.temp_dir / "certificates" / "developer.pem").exists())

        # Create a real extension
        dt_sdk.create("test_extension", self.temp_dir)

        # Check that the extension files exist
        self.assertTrue((self.temp_dir / "test_extension" / "test_extension" / "__main__.py").exists())
        self.assertTrue((self.temp_dir / "test_extension" / "setup.py").exists())
        self.assertTrue((self.temp_dir / "test_extension" / "activation.json").exists())
        self.assertTrue((self.temp_dir / "test_extension" / "extension" / "extension.yaml").exists())

        # Build the extension
        dt_sdk.build(
            self.temp_dir / "test_extension",
            self.temp_dir / "certificates" / "developer.pem",
            target_directory=None,
            extra_platforms=None,
            extra_index_url=None,
            find_links=None,
        )

        # Check that the built extension file exists
        self.assertTrue((self.temp_dir / "test_extension" / "extension" / "lib").exists())
        self.assertTrue((self.temp_dir / "test_extension" / "dist" / "custom_test-extension-0.0.1.zip").exists())



import logging
import threading
import time


class ThrottledHandler(logging.StreamHandler):
    def __init__(self, stream, log_repeat_interval=3600, cache_clean_interval=3600):
        super().__init__(stream)
        self.record_to_last_print_time = {}
        self.log_repeat_interval = log_repeat_interval
        self.record_cache_last_clean_timestamp = time.time()
        self.record_cache_clean_interval = cache_clean_interval
        self.record_cache_clean_lock = threading.Lock()

    def emit(self, record):
        current_time = time.time()
        if record.msg not in self.record_to_last_print_time.keys():
            super().emit(record)
            self.record_to_last_print_time[record.msg] = current_time
            return
        last_print_time = self.record_to_last_print_time[record.msg]
        if (last_print_time - current_time) > self.log_repeat_interval:
            self.record_to_last_print_time[record.msg] = current_time
            super().emit(record)
        self.cleanup(current_time)

    def cleanup(self, current_time):
        if (self.record_cache_last_clean_timestamp - current_time) > self.record_cache_clean_interval:
            with self.record_cache_clean_lock:
                keys_to_remove = []
                for record_message, last_print_time in self.record_to_last_print_time.items():
                    if (current_time - last_print_time) > self.log_repeat_interval:
                        keys_to_remove.append(record_message)
                for key in keys_to_remove:
                    del self.record_to_last_print_time[key]


class StrictThrottledHandler(logging.StreamHandler):
    def __init__(self, stream, log_repeat_interval=3600, cache_clean_interval=3600):
        super().__init__(stream)
        self.record_to_last_print_time = {}
        self.log_repeat_interval = log_repeat_interval
        self.record_cache_last_clean_timestamp = time.time()
        self.record_cache_clean_interval = cache_clean_interval
        self.record_cache_clean_lock = threading.Lock()

    def emit(self, record):
        current_time = time.time()
        log_identifier = f"{record.filename}:{record.lineno}"
        if log_identifier not in self.record_to_last_print_time.keys():
            super().emit(record)
            self.record_to_last_print_time[log_identifier] = current_time
            return
        last_print_time = self.record_to_last_print_time[log_identifier]
        if (last_print_time - current_time) > self.log_repeat_interval:
            self.record_to_last_print_time[log_identifier] = current_time
            super().emit(record)
        self.cleanup(current_time)

    def cleanup(self, current_time):
        if (self.record_cache_last_clean_timestamp - current_time) > self.record_cache_clean_interval:
            with self.record_cache_clean_lock:
                keys_to_remove = []
                for record_message, last_print_time in self.record_to_last_print_time.items():
                    if (current_time - last_print_time) > self.log_repeat_interval:
                        keys_to_remove.append(record_message)
                for key in keys_to_remove:
                    del self.record_to_last_print_time[key]
